@@ -301,6 +301,7 @@ setup_condition_monitor (GsmAutostartApp *app)
         key = NULL;
         res = parse_condition_string (app->priv->condition_string, &kind, &key);
         if (! res) {
+                g_free (key);
                 return;
         }
 
@@ -367,6 +368,8 @@ setup_condition_monitor (GsmAutostartApp *app)
                 disabled = TRUE;
         }
 
+        g_free (key);
+
         /* FIXME: cache the disabled value? */
 }
 
@@ -571,6 +574,8 @@ gsm_autostart_app_dispose (GObject *object)
         if (priv->condition_monitor) {
                 g_file_monitor_cancel (priv->condition_monitor);
         }
+
+        G_OBJECT_CLASS (gsm_autostart_app_parent_class)->dispose (object);
 }
 
 static gboolean
@@ -609,6 +614,7 @@ is_conditionally_disabled (GsmApp *app)
         key = NULL;
         res = parse_condition_string (priv->condition_string, &kind, &key);
         if (! res) {
+                g_free (key);
                 return TRUE;
         }
 
@@ -641,6 +647,8 @@ is_conditionally_disabled (GsmApp *app)
         /* Set initial condition */
         priv->condition = !disabled;
 
+        g_free (key);
+
         return disabled;
 }
 
@@ -935,11 +943,12 @@ gsm_autostart_app_restart (GsmApp  *app,
         GError  *local_error;
         gboolean res;
 
+        /* ignore stop errors - it is fine if it is already stopped */
         local_error = NULL;
         res = gsm_app_stop (app, &local_error);
         if (! res) {
-                g_propagate_error (error, local_error);
-                return FALSE;
+                g_debug ("GsmAutostartApp: Couldn't stop app: %s", local_error->message);
+                g_error_free (local_error);
         }
 
         res = gsm_app_start (app, &local_error);
