@@ -639,6 +639,8 @@ gsm_xsmp_client_finalize (GObject *object)
         gsm_xsmp_client_disconnect (client);
 
         g_free (client->priv->description);
+        g_ptr_array_foreach (client->priv->props, (GFunc)SmFreeProperty, NULL);
+        g_ptr_array_free (client->priv->props, TRUE);
 
         G_OBJECT_CLASS (gsm_xsmp_client_parent_class)->finalize (object);
 }
@@ -861,6 +863,8 @@ register_client_callback (SmsConn    conn,
 
         gsm_client_set_status (GSM_CLIENT (client), GSM_CLIENT_REGISTERED);
 
+        g_free (id);
+
         return TRUE;
 }
 
