@@ -25,6 +25,8 @@
 #include <signal.h>
 #include <stdlib.h>
 #include <string.h>
+#include <unistd.h>
+#include <errno.h>
 
 #include <glib/gi18n.h>
 #include <glib/goption.h>
@@ -46,6 +48,7 @@
 
 #define GSM_GCONF_DEFAULT_SESSION_KEY           "/desktop/gnome/session/default-session"
 #define GSM_GCONF_REQUIRED_COMPONENTS_DIRECTORY "/desktop/gnome/session/required-components"
+#define GSM_GCONF_REQUIRED_COMPONENTS_KEY       "/desktop/gnome/session/required-components"
 
 #define GSM_DBUS_NAME "org.gnome.SessionManager"
 
@@ -53,18 +56,9 @@
 
 static gboolean failsafe = FALSE;
 static gboolean show_version = FALSE;
-static char **override_autostart_dirs = NULL;
 /* FIXME: turn this off closer to release */
 static gboolean debug = TRUE;
 
-static GOptionEntry entries[] = {
-        { "autostart", 'a', 0, G_OPTION_ARG_STRING_ARRAY, &override_autostart_dirs, N_("Override standard autostart directories"), NULL },
-        { "debug", 0, 0, G_OPTION_ARG_NONE, &debug, N_("Enable debugging code"), NULL },
-        { "failsafe", 'f', 0, G_OPTION_ARG_NONE, &failsafe, N_("Do not load user-specified applications"), NULL },
-        { "version", 0, 0, G_OPTION_ARG_NONE, &show_version, N_("Version of this application"), NULL },
-        { NULL, 0, 0, 0, NULL, NULL, NULL }
-};
-
 static void
 on_bus_name_lost (DBusGProxy *bus_proxy,
                   const char *name,
@@ -169,88 +163,127 @@ acquire_name (void)
         return TRUE;
 }
 
+static char *
+find_desktop_file_for_app_name (const char *name,
+                                char      **autostart_dirs)
+{
+        char     *app_path;
+        char    **app_dirs;
+        GKeyFile *key_file;
+        char     *desktop_file;
+        int       i;
+
+        app_path = NULL;
+
+        app_dirs = gsm_util_get_app_dirs ();
+
+        key_file = g_key_file_new ();
+
+        desktop_file = g_strdup_printf ("%s.desktop", name);
+
+        g_debug ("main: Looking for file '%s'", desktop_file);
+
+        for (i = 0; app_dirs[i] != NULL; i++) {
+                g_debug ("main: Looking in '%s'", app_dirs[i]);
+        }
+
+        g_key_file_load_from_dirs (key_file,
+                                   desktop_file,
+                                   (const char **) app_dirs,
+                                   &app_path,
+                                   G_KEY_FILE_NONE,
+                                   NULL);
+
+        if (app_path != NULL) {
+                g_debug ("main: found in XDG app dirs: '%s'", app_path);
+        }
+
+        if (app_path == NULL && autostart_dirs != NULL) {
+                g_key_file_load_from_dirs (key_file,
+                                           desktop_file,
+                                           (const char **) autostart_dirs,
+                                           &app_path,
+                                           G_KEY_FILE_NONE,
+                                           NULL);
+                if (app_path != NULL) {
+                        g_debug ("main: found in autostart dirs: '%s'", app_path);
+                }
+
+        }
+
+        /* look for gnome vender prefix */
+        if (app_path == NULL) {
+                g_free (desktop_file);
+                desktop_file = g_strdup_printf ("gnome-%s.desktop", name);
+
+                g_key_file_load_from_dirs (key_file,
+                                           desktop_file,
+                                           (const char **) app_dirs,
+                                           &app_path,
+                                           G_KEY_FILE_NONE,
+                                           NULL);
+                if (app_path != NULL) {
+                        g_debug ("main: found in XDG app dirs: '%s'", app_path);
+                }
+        }
+
+        if (app_path == NULL && autostart_dirs != NULL) {
+                g_key_file_load_from_dirs (key_file,
+                                           desktop_file,
+                                           (const char **) autostart_dirs,
+                                           &app_path,
+                                           G_KEY_FILE_NONE,
+                                           NULL);
+                if (app_path != NULL) {
+                        g_debug ("main: found in autostart dirs: '%s'", app_path);
+                }
+        }
+
+        g_free (desktop_file);
+        g_key_file_free (key_file);
+
+        g_strfreev (app_dirs);
+
+        return app_path;
+}
+
 static void
 append_default_apps (GsmManager *manager,
+                     const char *default_session_key,
                      char      **autostart_dirs)
 {
         GSList      *default_apps;
         GSList      *a;
-        char       **app_dirs;
         GConfClient *client;
 
         g_debug ("main: *** Adding default apps");
 
-        app_dirs = gsm_util_get_app_dirs ();
+        g_assert (default_session_key != NULL);
+        g_assert (autostart_dirs != NULL);
 
         client = gconf_client_get_default ();
         default_apps = gconf_client_get_list (client,
-                                              GSM_GCONF_DEFAULT_SESSION_KEY,
+                                              default_session_key,
                                               GCONF_VALUE_STRING,
                                               NULL);
         g_object_unref (client);
 
         for (a = default_apps; a; a = a->next) {
-                GKeyFile *key_file;
-                char     *app_path = NULL;
-                char     *desktop_file;
-
-                key_file = g_key_file_new ();
+                char *app_path;
 
                 if (IS_STRING_EMPTY ((char *)a->data)) {
                         continue;
                 }
 
-                desktop_file = g_strdup_printf ("%s.desktop", (char *) a->data);
-                g_key_file_load_from_dirs (key_file,
-                                           desktop_file,
-                                           (const char**) app_dirs,
-                                           &app_path,
-                                           G_KEY_FILE_NONE,
-                                           NULL);
-
-                if (app_path == NULL) {
-                        g_key_file_load_from_dirs (key_file,
-                                                   desktop_file,
-                                                   (const char**) autostart_dirs,
-                                                   &app_path,
-                                                   G_KEY_FILE_NONE,
-                                                   NULL);
-                }
-
-                /* look for gnome vender prefix */
-                if (app_path == NULL) {
-                        g_free (desktop_file);
-                        desktop_file = g_strdup_printf ("gnome-%s.desktop", (char *) a->data);
-
-                        g_key_file_load_from_dirs (key_file,
-                                                   desktop_file,
-                                                   (const char**) app_dirs,
-                                                   &app_path,
-                                                   G_KEY_FILE_NONE,
-                                                   NULL);
-                }
-
-                if (app_path == NULL) {
-                        g_key_file_load_from_dirs (key_file,
-                                                   desktop_file,
-                                                   (const char**) autostart_dirs,
-                                                   &app_path,
-                                                   G_KEY_FILE_NONE,
-                                                   NULL);
-                }
-
+                app_path = find_desktop_file_for_app_name (a->data, autostart_dirs);
                 if (app_path != NULL) {
                         gsm_manager_add_autostart_app (manager, app_path, NULL);
                         g_free (app_path);
                 }
-
-                g_free (desktop_file);
-                g_key_file_free (key_file);
         }
 
         g_slist_foreach (default_apps, (GFunc) g_free, NULL);
         g_slist_free (default_apps);
-        g_strfreev (app_dirs);
 }
 
 static void
@@ -280,46 +313,61 @@ append_required_apps (GsmManager *manager)
         GConfClient *client;
 
         client = gconf_client_get_default ();
-        required_components = gconf_client_all_entries (client,
-                                                        GSM_GCONF_REQUIRED_COMPONENTS_DIRECTORY,
-                                                        NULL);
-        g_object_unref (client);
+        required_components = gconf_client_get_list (client,
+                                                     GSM_GCONF_REQUIRED_COMPONENTS_KEY,
+                                                     GCONF_VALUE_STRING,
+                                                     NULL);
 
         for (r = required_components; r != NULL; r = r->next) {
-                GConfEntry *entry;
-                const char *default_provider;
-                const char *service;
+                char       *path;
+                char       *default_provider;
+                const char *component;
 
-                entry = (GConfEntry *) r->data;
-
-                service = strrchr (entry->key, '/');
-                if (service == NULL) {
-                        continue;
-                }
-                service++;
-
-                default_provider = gconf_value_get_string (entry->value);
-                if (default_provider == NULL) {
+                if (IS_STRING_EMPTY ((char *)r->data)) {
                         continue;
                 }
 
-                gsm_manager_add_autostart_app (manager, default_provider, service);
+                component = r->data;
 
-                gconf_entry_free (entry);
+                path = g_strdup_printf ("%s/%s",
+                                        GSM_GCONF_REQUIRED_COMPONENTS_DIRECTORY,
+                                        component);
+
+                default_provider = gconf_client_get_string (client, path, NULL);
+                if (default_provider != NULL) {
+                        char *app_path;
+
+                        app_path = find_desktop_file_for_app_name (default_provider, NULL);
+                        if (app_path != NULL) {
+                                gsm_manager_add_autostart_app (manager, app_path, component);
+                        } else {
+                                g_warning ("Unable to find provider '%s' of required component '%s'",
+                                           default_provider,
+                                           component);
+                        }
+                        g_free (app_path);
+                }
+
+                g_free (default_provider);
+                g_free (path);
         }
 
+        g_slist_foreach (required_components, (GFunc)g_free, NULL);
         g_slist_free (required_components);
+
+        g_object_unref (client);
 }
 
 static void
-load_standard_apps (GsmManager *manager)
+load_standard_apps (GsmManager *manager,
+                    const char *default_session_key)
 {
         char **autostart_dirs;
         int    i;
 
         autostart_dirs = gsm_util_get_autostart_dirs ();
 
-        append_default_apps (manager, autostart_dirs);
+        append_default_apps (manager, default_session_key, autostart_dirs);
 
         if (failsafe) {
                 goto out;
@@ -340,7 +388,8 @@ load_standard_apps (GsmManager *manager)
 }
 
 static void
-load_override_apps (GsmManager *manager)
+load_override_apps (GsmManager *manager,
+                    char      **override_autostart_dirs)
 {
         int i;
         for (i = 0; override_autostart_dirs[i]; i++) {
@@ -390,6 +439,46 @@ signal_cb (int      signo,
         return ret;
 }
 
+static gboolean
+require_dbus_session (int      argc,
+                      char   **argv,
+                      GError **error)
+{
+        char **new_argv;
+        int    i;
+
+        if (g_getenv ("DBUS_SESSION_BUS_ADDRESS"))
+                return TRUE;
+
+        /* Just a sanity check to prevent infinite recursion if
+         * dbus-launch fails to set DBUS_SESSION_BUS_ADDRESS 
+         */
+        g_return_val_if_fail (!g_str_has_prefix (argv[0], "dbus-launch"),
+                              TRUE);
+
+        /* +2 for our new arguments, +1 for NULL */
+        new_argv = g_malloc (argc + 3 * sizeof (*argv));
+        
+        new_argv[0] = "dbus-launch";
+        new_argv[1] = "--exit-with-session";
+        new_argv[2] = argv[0];
+        for (i = 1; i < argc - 1; i++)
+                new_argv[i + 2] = argv[i];
+        new_argv[i + 2] = NULL;
+        
+        if (!execvp ("dbus-launch", new_argv)) {
+                g_set_error (error, 
+                             G_SPAWN_ERROR,
+                             G_SPAWN_ERROR_FAILED,
+                             "No session bus and could not exec dbus-launch: %s",
+                             g_strerror (errno));
+                return FALSE;
+        }
+
+        /* Should not be reached */
+        return TRUE;
+}
+
 int
 main (int argc, char **argv)
 {
@@ -400,6 +489,21 @@ main (int argc, char **argv)
         GsmStore         *client_store;
         GsmXsmpServer    *xsmp_server;
         GdmSignalHandler *signal_handler;
+        static char     **override_autostart_dirs = NULL;
+        static char      *default_session_key = NULL;
+        static GOptionEntry entries[] = {
+                { "autostart", 'a', 0, G_OPTION_ARG_STRING_ARRAY, &override_autostart_dirs, N_("Override standard autostart directories"), NULL },
+                { "default-session-key", 0, 0, G_OPTION_ARG_STRING, &default_session_key, N_("GConf key used to lookup default session"), NULL },
+                { "debug", 0, 0, G_OPTION_ARG_NONE, &debug, N_("Enable debugging code"), NULL },
+                { "failsafe", 'f', 0, G_OPTION_ARG_NONE, &failsafe, N_("Do not load user-specified applications"), NULL },
+                { "version", 0, 0, G_OPTION_ARG_NONE, &show_version, N_("Version of this application"), NULL },
+                { NULL, 0, 0, 0, NULL, NULL, NULL }
+        };
+
+        /* Make sure that we have a session bus */
+        if (!require_dbus_session (argc, argv, &error)) {
+                gsm_util_init_error (TRUE, "%s", error->message);
+        }
 
         bindtextdomain (GETTEXT_PACKAGE, LOCALE_DIR);
         bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
@@ -444,6 +548,11 @@ main (int argc, char **argv)
         gsm_util_setenv ("DISPLAY", display_str);
         g_free (display_str);
 
+        /* Some third-party programs rely on GNOME_DESKTOP_SESSION_ID to
+         * detect if GNOME is running. We keep this for compatibility reasons.
+         */
+        gsm_util_setenv ("GNOME_DESKTOP_SESSION_ID", "this-is-deprecated");
+
         client_store = gsm_store_new ();
 
 
@@ -460,9 +569,13 @@ main (int argc, char **argv)
 
         manager = gsm_manager_new (client_store, failsafe);
         if (override_autostart_dirs != NULL) {
-                load_override_apps (manager);
+                load_override_apps (manager, override_autostart_dirs);
         } else {
-                load_standard_apps (manager);
+                if (! IS_STRING_EMPTY (default_session_key)) {
+                        load_standard_apps (manager, default_session_key);
+                } else {
+                        load_standard_apps (manager, GSM_GCONF_DEFAULT_SESSION_KEY);
+                }
         }
 
         gsm_xsmp_server_start (xsmp_server);
